# -*- coding: utf-8 -*-
import logging
import psycopg2
import requests
import time
import werkzeug
import uuid
import hmac
from hashlib import sha1

from odoo import api, fields, models
from odoo.addons.auth_signup.models.res_partner import now
from odoo.addons.saas_worker.auth_util import SaasAuth
from odoo.exceptions import AccessDenied, ValidationError
from odoo.osv import expression
from odoo.release import version
from odoo.tools import conditional, email_re
from odoo.tools.parse_version import parse_version
from odoo.tools.translate import _
from odoo.addons.saas_worker.compat import consteq, check_identity
from odoo.addons.saas_worker.const import OAUTH_BASE_URL, TIMEOUT
from odoo.addons.saas_worker import metabase
from odoo.addons.saas_worker import trials
from odoo.addons.saas_worker.util import version_gte
from odoo.http import request

_logger = logging.getLogger(__name__)


def _add_saas_trial_oauth_uid_ir_model_data(cr):
    # Add the `res.users.oauth_uid` field XML ID for the `saas_trial` module if it doesn't exist.
    # This is to prevent the deletion of the column `oauth_uid` of the table `res_users`
    # upon the uninstall of the auth_oauth module.
    # This only happens for the databases using the OAuth Odoo.com authentication before the
    # introduction of the field `oauth_uid` in the module `saas_trial`.
    # For newer databases, the issue doesn't occur, as this xml_id correctly exists.
    # This can be deleted as soon as all databases on the saas have this xml id,
    # meaning they all had at some point an update of the `saas_trial` module
    cr.execute("SELECT * FROM ir_model_data WHERE module = 'saas_trial' AND name = 'field_res_users_oauth_uid'")
    if not cr.rowcount:
        cr.execute("SELECT id from ir_model_fields where model = 'res.users' AND name = 'oauth_uid'")
        field = cr.fetchone()
        if field:
            columns = ("name", "module", "model", "res_id")
            values = ("'field_res_users_oauth_uid'", "'saas_trial'", "'ir.model.fields'", "%s")
            query = "INSERT INTO ir_model_data (%s) VALUES (%s)" % (",".join(columns), ", ".join(values))
            cr.execute(query, (field[0],))


class ResUsers(models.Model):
    _inherit = 'res.users'

    @api.model_cr
    def init(self):
        _add_saas_trial_oauth_uid_ir_model_data(self._cr)

    def _get_odoo_com_uid(self):
        """
            A user is considered to use the Odoo.com server-to-server authentication only if oauth_uid is set and oauth_provider_id is not.
        """
        for user in self:
            user.odoo_com_uid = user.oauth_uid if 'oauth_provider_id' not in self._fields or not user.oauth_provider_id else False

    def _set_odoo_com_uid(self):
        for user in self:
            user.oauth_uid = user.odoo_com_uid
            if 'oauth_provider_id' in self._fields:
                user.oauth_provider_id = False

    def _search_odoo_com_uid(self, operator, operand):
        domain = [('oauth_uid', operator, operand)]
        if 'oauth_provider_id' in self._fields:
            domain = expression.AND([domain, [('oauth_provider_id', '=', False)]])
        return domain

    oauth_uid = fields.Char('OAuth User ID', help="Oauth Provider user_id", copy=False)
    odoo_com_uid = fields.Char(compute='_get_odoo_com_uid', inverse='_set_odoo_com_uid', search='_search_odoo_com_uid', string='Odoo.com User ID')

    @api.model
    def _assign_login_email(self, values):
        if values.get('login'):
            login = values['login']
            lp, at, hn = login.partition('@')
            login = lp + at + hn.lower()
            if login != 'anonymous' and not login.startswith(('embedded-', '__deleted_user_')) and not email_re.match(login):
                # if demo data are loaded
                if self.env['ir.module.module'].search_count([('demo', '=', False)]) == 0:
                    values['active'] = False
                elif values.get('active', True):
                    raise ValidationError(_('Login must be a valid email address : %s') % login)
            else:
                values['login'] = values['email'] = login

    # overridden to assign login to email
    @api.model
    def create(self, values):
        self._assign_login_email(values)
        res = super(ResUsers, self).create(values)
        if not self.env.context.get('no_update_notification') and not res.share:
            self.env['publisher_warranty.contract'].sudo().with_context(force_ping=True).update_notification(cron_mode=True)
        return res

    @api.multi
    def write(self, values):
        self._assign_login_email(values)
        if ('login' in values or 'password' in values) and not self._context.get('saas_force_oauth'):
            odoo_com_users = [user for user in self if user.odoo_com_uid]
            if odoo_com_users:
                raise ValidationError(_(
                    "You cannot change the credentials of the user %s as they use Odoo.com to sign in.\n"
                    "If you want to change the password, you can send the reset password instructions"
                    "by clicking on the according button in the user form.\n"
                    "If you want to change the Odoo.com login, please contact the support (https://www.odoo.com/help)."
                ) % odoo_com_users[0].login)
        group_system = self.env.ref('base.group_system')
        had_group_system = {u.id for u in self if group_system.id in u.groups_id.ids}
        res = super(ResUsers, self).write(values)

        if not self.env.context.get('no_update_notification'):
            should_ping = bool(set(values) & {'login', 'password', 'active'})
            # conditions before this line are enough to force the ping
            force_ping = should_ping
            # conditions after this line are doing a soft ping (changing should_ping but not force_ping)
            have_group_system = {u.id for u in self if group_system.id in u.groups_id.ids}
            should_ping |= had_group_system != have_group_system

            if should_ping:
                self.env['publisher_warranty.contract'].sudo().with_context(force_ping=force_ping).update_notification(cron_mode=True)

        return res

    @api.multi
    @conditional(
        version_gte("16.0"),
        check_identity,
    )
    def preference_change_password(self):
        user = self.env.user.sudo()
        odoo_provider = self.env.ref('auth_oauth.provider_openerp', raise_if_not_found=False)
        if not user.odoo_com_uid and odoo_provider and user.oauth_provider_id and user.oauth_provider_id.enabled and user.oauth_provider_id == odoo_provider:
            return {
                'target': 'self',
                'type': 'ir.actions.act_url',
                'url': '%s/web#action=change_password' % (OAUTH_BASE_URL)
            }
        else:
            return super(ResUsers, self).preference_change_password()

    @api.model
    def change_password(self, old_passwd, new_passwd):
        user = self.env.user.sudo()
        if user.odoo_com_uid:
            if not old_passwd:
                # empty passwords disallowed for obvious security reasons
                raise AccessDenied()
            self._check_credentials(old_passwd, {'interactive': True})
            url = '%s/odoo-enterprise/trial-login-change-password' % OAUTH_BASE_URL
            data = {'params': {'user_id': user.odoo_com_uid, 'password': old_passwd, 'new_password': new_passwd}}
            result = requests.post(url, json=data, auth=SaasAuth(), timeout=TIMEOUT).json()
            if result.get('result'):
                # Trigger a change on the user to empty the check_credentials cache
                user.write({'active': user.active})
            else:
                raise ValidationError(_('The attempt to change your Odoo.com password failed. Please change your password directly on accounts.odoo.com'))
            return True
        else:
            return super(ResUsers, self).change_password(old_passwd, new_passwd)

    @api.multi
    def action_reset_password(self):
        create_mode = bool(self.env.context.get('create_user'))
        odoo_com_linked_users = self.browse()
        odoo_com_existing_users = self.browse()
        internal_users = self.browse()
        for user in self:
            if user.odoo_com_uid:
                odoo_com_linked_users |= user
            elif (not create_mode) and (not user.partner_share) and user.partner_id._user_exists_odoocom():
                odoo_com_existing_users |= user
            else:
                internal_users |= user
        super(ResUsers, internal_users).action_reset_password()

        odoocom_passwords_to_reset = []
        db_uuid = self.env['ir.config_parameter'].get_param('database.uuid', default='')
        redirect_url = f'/my/databases/connect/{db_uuid}' if db_uuid else '/my/databases'
        for user in odoo_com_linked_users:
            odoocom_passwords_to_reset.append({'login': user.login, 'saas_trial_reset': redirect_url})

        # OAuth linking via set_oauth_uid needs a valid signup token
        create_mode = bool(self.env.context.get('create_user'))
        expiration = False if create_mode else now(days=+1)
        odoo_com_existing_users.partner_id.signup_prepare(signup_type="reset", expiration=expiration)
        for user in odoo_com_existing_users:
            odoocom_passwords_to_reset.append({'login': user.login, 'saas_trial_reset': user.signup_url})

        if odoocom_passwords_to_reset:
            requests.post(f'{OAUTH_BASE_URL}/openerp_enterprise/reset_passwords', json={'params': {
                'passwords_to_reset': odoocom_passwords_to_reset,
            }}, auth=SaasAuth(), timeout=TIMEOUT)

    @api.model
    def _check_credentials(self, password, env={}):
        try:
            super(ResUsers, self)._check_credentials(password, env)
        except AccessDenied as access_denied:
            try:
                clauses, clauses_args = ['id = %s', 'oauth_uid is not NULL', 'active=TRUE'], [self.env.user.id]
                if 'oauth_provider_id' in self._fields:
                    clauses.append('(oauth_provider_id IS NULL)')
                self.env.cr.execute('SELECT oauth_uid FROM res_users WHERE %s' % ' AND '.join(clauses), tuple(clauses_args))
                res = self.env.cr.fetchone()
                if res:
                    spoof_params = {
                        'user_id': res[0],
                        'password': password,
                    }

                    interactive_login = bool(env.get('interactive', True))

                    try:  # don't crash for this
                        spoof_params.update({
                            'reason': 'RPC' if not interactive_login else '',
                            'action': 'login',
                            'remote_uid': self.env.user.id,  # remote from PoV of accounts.odoo.com
                            'remote_login': self.env.user.login,
                            'dbuuid': self.env['ir.config_parameter'].sudo().get_param('database.uuid'),
                            })
                        if request:
                            spoof_params['ip'] = request.httprequest.environ['REMOTE_ADDR']
                    except Exception:
                        pass
                    if 'interactive' in env:
                        spoof_params['interactive'] = env['interactive']
                    params = {'params': spoof_params}
                    result = requests.post(
                        '%s/odoo-enterprise/trial-login' % OAUTH_BASE_URL, json=params,
                        auth=SaasAuth(), timeout=TIMEOUT,
                    ).json().get('result')
                    if result and result.get('is_staff_restricted') is not True:
                        if request:
                            request.session.impostor_info = result.get('impostor_info', {})
                        if not result.get('support') and interactive_login:
                            totp_enabled = result.get('totp_enabled') and 'totp_enabled' in self.env.user._fields
                            if totp_enabled:
                                request.session.odoo_com_totp_check = result['totp_check_token']

                        if not interactive_login:
                            try:
                                msg = 'Non-interactive(RPC) authentication from %s using the odoo.com account %s' % (
                                    spoof_params.get('ip') or 'an unknown IP', self.env.user.login
                                )
                                ir_log = self.env['ir.logging']
                                ir_log.sudo().create({
                                    'name': 'RPC Authentication',
                                    'type': 'server',
                                    'dbname': self.env.cr.dbname,
                                    'path': __name__,
                                    'func': 'check_credentials',
                                    'line': '0',
                                    'level': 'INFO',
                                    'message': msg,
                                })
                            except Exception:
                                _logger.exception(
                                    'Unable to log in ir.logging a remote non-interactive authentication: %s', msg
                                )

                        # support: only allow directly access for RPC and with api-key (2FA enabled)
                        if result.get('support') and (interactive_login or not result.get('totp_enabled')):
                            _logger.exception(
                                'Unable to authenticate user "%s" using "%s" employee magic password (need 2FA)' % (
                                    self.env.user.id, result['impostor_info']['login'],
                                )
                            )
                        else:
                            return  # access granted
            except Exception:
                _logger.warning('Unable to authenticate user %s using Odoo.com', self.env.user.id, exc_info=True)
            raise access_denied

    def _mfa_type(self):
        if request and request.session.odoo_com_totp_check:
            return 'totp'
        return super()._mfa_type()

    def _send_totp_mail_code(self):
        # For staging databases the outgoing mail server is deactivated, so TOTP by email isn't possible
        try:
            return super()._send_totp_mail_code()
        except Exception:
            _logger.exception("2fa by email failed")
            raise AccessDenied(_("Two-factor authentication by email is not available on this database, "
                "you may want to set up TOTP via an authentication App instead.\n"
                "Reminder: outgoing emails are disabled by default on databases duplicated for testing purposes"))

    def _totp_check(self, code):
        if not request.session.odoo_com_totp_check:
            return super()._totp_check(code)
        user = self.sudo()
        url = '%s/odoo-enterprise/trial-totp-check' % OAUTH_BASE_URL
        data = {'params': {
            'user_id': user.odoo_com_uid,
            'code': code,
            'check_token': request.session.odoo_com_totp_check,
        }}
        try:
            result = requests.post(url, json=data, auth=SaasAuth(), timeout=TIMEOUT).json()
        except Exception:
            result = {}
        check = result.get('result', {}).get('result')
        if check:
            request.session.pop('odoo_com_totp_check')
            return
        error = result.get('result', {}).get('error') or _("The odoo.com two-factor authentication failed.")
        raise AccessDenied(error)

    @api.model
    def auth_oauth(self, provider, params):
        # This method is supposed to be called only for databases having auth_oauth installed.
        # For the databases using the Odoo.com OAuth provider as authentication server
        # it allows to not override the user access_token with a new access token when
        # support employees access the database as imposotors. Instead it uses the access token
        # already stored.
        db, login, _ = super(ResUsers, self).auth_oauth(provider, params)
        return db, login, params['access_token']

    @api.model
    def _auth_oauth_signin(self, provider, validation, params):
        # This method is supposed to be called only for databases having auth_oauth installed.
        # See the comment of the above `auth_oauth` method for more information.
        if validation.get('support'):
            oauth_uid = validation['user_id']
            users = self.search_read([("oauth_uid", "=", oauth_uid), ('oauth_provider_id', '=', provider)], ['oauth_access_token'])
            if users and users[0]['oauth_access_token']:
                params['access_token'] = users[0]['oauth_access_token']
        return super(ResUsers, self)._auth_oauth_signin(provider, validation, params)

    @classmethod
    def _login(cls, db, login, password, user_agent_env={}):
        uid = super(ResUsers, cls)._login(db, login, password, user_agent_env=user_agent_env)
        cls.update_cnx_date(db, uid)
        return uid

    @classmethod
    def update_cnx_date(cls, db, uid):
        if uid:
            try:
                mdb = metabase.get(db)
                if mdb:
                    metabase.update(mdb, last_cnx_date="now() AT TIME ZONE 'UTC'")
                    trials.bump_sliding_expiration(mdb)
            except psycopg2.Error:
                pass
        return True

    def _get_secret_signature(self, expiration=7 * 24 * 60 * 60, max_ts=None):
        self.ensure_one()
        ICP = self.env['ir.config_parameter'].sudo()
        key = ICP.get_param('database.secret')
        if not key:
            key = str(uuid.uuid4())
            ICP.set_param('database.secret', key)

        if max_ts is None:
            max_ts = int(time.time() + expiration)

        data = (u'SaaS>PaaS>Tasse-%s.%s>%s' % (self.id, self.login, max_ts)).encode('utf-8')
        h = hmac.new(key.encode('ascii'), data, sha1)
        return '%so%s' % (h.hexdigest(), max_ts)

    def _check_secret_signature(self, pretender):
        now = int(time.time())
        max_ts = pretender.partition("o")[2]
        try:
            max_ts = int(max_ts)
        except ValueError:
            return False

        return now < max_ts and consteq(self._get_secret_signature(max_ts=max_ts), pretender)

    def _can_import_remote_urls(self):
        # Disable remote URL imports - cfr feature in base_import for v12
        # TODO: selectively enable the feature on demand, e.g. based on the presence of a special
        #       ir.config.parameter entry
        return False

class ChangePasswordWizard(models.TransientModel):
    """ A wizard to manage the change of users' passwords. """
    _inherit = "change.password.wizard"

    @api.multi
    def change_password_button(self):
        self.ensure_one()
        if any(self.user_ids.mapped('user_id.odoo_com_uid')) and not self._context.get('saas_force_oauth'):
            compose_form = self.env.ref('saas_trial.change_password_wizard_view', False)
            # the view may not exist (yet)
            if compose_form:
                ctx = self._context.copy()
                ctx.update(saas_force_oauth=True)
                return {
                    'type': 'ir.actions.act_window',
                    'view_type': 'form',
                    'view_mode': 'form',
                    'res_model': 'change.password.wizard',
                    'res_id': self.id,
                    'views': [(compose_form.id, 'form')],
                    'view_id': compose_form.id,
                    'target': 'new',
                    'context': ctx,
                }

        return super(ChangePasswordWizard, self).change_password_button()
